@@ -1,6 +1,7 @@
 # init.d -- lintian check script -*- perl -*-
 
-# Copyright (C) 1998 Christian Schwarz
+# Copyright © 1998 Christian Schwarz
+# Copyright © 2016-2020 Chris Lamb <lamby@debian.org>
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -19,15 +20,21 @@
 # MA 02110-1301, USA.
 
 package Lintian::init_d;
-use strict;
+
+use v5.20;
 use warnings;
+use utf8;
 use autodie;
 
 use File::Basename qw(dirname);
 use List::MoreUtils qw(any none);
 
 use Lintian::Data;
-use Lintian::Tags qw(tag);
+
+use Moo;
+use namespace::clean;
+
+with 'Lintian::Check';
 
 # A list of valid LSB keywords.  The value is 0 if optional and 1 if required.
 my %lsb_keywords = (
@@ -76,33 +83,44 @@ my $OPTS_R = qr/-\S+\s*/;
 my $ACTION_R = qr/\w+/;
 my $EXCLUDE_R = qr/if\s+\[\s+-x\s+\S*update-rc\.d/;
 
-sub run {
-    my (undef, undef, $info) = @_;
-    my $initd_dir = $info->index_resolved_path('etc/init.d/');
-    my $postinst = $info->control_index('postinst');
-    my $preinst = $info->control_index('preinst');
-    my $postrm = $info->control_index('postrm');
-    my $prerm = $info->control_index('prerm');
+my $ALWAYS_START
+  = qr/^\s*#*\s*(?:[A-Z]_)?(?:ENABLED|DISABLED|[A-Z]*RUN|(?:NO_)?START)=/;
+
+sub installable {
+    my ($self) = @_;
+
+    my $pkg = $self->processable->name;
+    my $processable = $self->processable;
+
+    my $initd_dir = $processable->installed->resolve_path('etc/init.d/');
+    my $postinst = $processable->control->lookup('postinst');
+    my $preinst = $processable->control->lookup('preinst');
+    my $postrm = $processable->control->lookup('postrm');
+    my $prerm = $processable->control->lookup('prerm');
 
     my (%initd_postinst, %initd_postrm);
 
+    # These will never be regular initscripts. (see #918459, #933383
+    # and #941140 etc.)
+    return if $pkg eq 'initscripts' or $pkg eq 'sysvinit';
+
     # read postinst control file
     if ($postinst and $postinst->is_file and $postinst->is_open_ok) {
-        my $fd = $postinst->open;
+        open(my $fd, '<', $postinst->unpacked_path);
         while (<$fd>) {
-            next if /$EXCLUDE_R/o;
-            s/\#.*$//o;
+            next if /$EXCLUDE_R/;
+            s/\#.*$//;
             next unless /^(?:.+;|^\s*system[\s\(\']+)?\s*update-rc\.d\s+
-            (?:$OPTS_R)*($INITD_NAME_REGEX)\s+($ACTION_R)/xo;
+            (?:$OPTS_R)*($INITD_NAME_REGEX)\s+($ACTION_R)/x;
             my ($name,$opt) = ($1,$2);
             next if $opt eq 'remove';
             if ($initd_postinst{$name}++ == 1) {
-                tag 'duplicate-updaterc.d-calls-in-postinst', $name;
+                $self->tag('duplicate-updaterc.d-calls-in-postinst', $name);
                 next;
             }
-            unless (m,>\s*/dev/null,o) {
-                tag 'output-of-updaterc.d-not-redirected-to-dev-null',
-                  "$name postinst";
+            unless (m,>\s*/dev/null,) {
+                $self->tag('output-of-updaterc.d-not-redirected-to-dev-null',
+                    "$name postinst");
             }
         }
         close($fd);
@@ -110,34 +128,34 @@ sub run {
 
     # read preinst control file
     if ($preinst and $preinst->is_file and $preinst->is_open_ok) {
-        my $fd = $preinst->open;
+        open(my $fd, '<', $preinst->unpacked_path);
         while (<$fd>) {
-            next if /$EXCLUDE_R/o;
-            s/\#.*$//o;
+            next if /$EXCLUDE_R/;
+            s/\#.*$//;
             next unless m/update-rc\.d \s+
                        (?:$OPTS_R)*($INITD_NAME_REGEX) \s+
-                       ($ACTION_R)/ox;
+                       ($ACTION_R)/x;
             my ($name,$opt) = ($1,$2);
             next if $opt eq 'remove';
-            tag 'preinst-calls-updaterc.d', $name;
+            $self->tag('preinst-calls-updaterc.d', $name);
         }
         close($fd);
     }
 
     # read postrm control file
     if ($postrm and $postrm->is_file and $postrm->is_open_ok) {
-        my $fd = $postrm->open;
+        open(my $fd, '<', $postrm->unpacked_path);
         while (<$fd>) {
-            next if /$EXCLUDE_R/o;
-            s/\#.*$//o;
-            next unless m/update-rc\.d\s+($OPTS_R)*($INITD_NAME_REGEX)/o;
+            next if /$EXCLUDE_R/;
+            s/\#.*$//;
+            next unless m/update-rc\.d\s+($OPTS_R)*($INITD_NAME_REGEX)/;
             if ($initd_postrm{$2}++ == 1) {
-                tag 'duplicate-updaterc.d-calls-in-postrm', $2;
+                $self->tag('duplicate-updaterc.d-calls-in-postrm', $2);
                 next;
             }
-            unless (m,>\s*/dev/null,o) {
-                tag 'output-of-updaterc.d-not-redirected-to-dev-null',
-                  "$2 postrm";
+            unless (m,>\s*/dev/null,) {
+                $self->tag('output-of-updaterc.d-not-redirected-to-dev-null',
+                    "$2 postrm");
             }
         }
         close($fd);
@@ -145,12 +163,12 @@ sub run {
 
     # read prerm control file
     if ($prerm and $prerm->is_file and $prerm->is_open_ok) {
-        my $fd = $prerm->open;
+        open(my $fd, '<', $prerm->unpacked_path);
         while (<$fd>) {
-            next if /$EXCLUDE_R/o;
-            s/\#.*$//o;
-            next unless m/update-rc\.d\s+($OPTS_R)*($INITD_NAME_REGEX)/o;
-            tag 'prerm-calls-updaterc.d', $2;
+            next if /$EXCLUDE_R/;
+            s/\#.*$//;
+            next unless m/update-rc\.d\s+($OPTS_R)*($INITD_NAME_REGEX)/;
+            $self->tag('prerm-calls-updaterc.d', $2);
         }
         close($fd);
     }
@@ -160,12 +178,13 @@ sub run {
         if ($initd_postrm{$_}) {
             delete $initd_postrm{$_};
         } else {
-            tag 'postrm-does-not-call-updaterc.d-for-init.d-script',
-              "etc/init.d/$_";
+            $self->tag('postrm-does-not-call-updaterc.d-for-init.d-script',
+                "etc/init.d/$_");
         }
     }
     for (keys %initd_postrm) {
-        tag 'postrm-contains-additional-updaterc.d-calls', "etc/init.d/$_";
+        $self->tag('postrm-contains-additional-updaterc.d-calls',
+            "etc/init.d/$_");
     }
 
     foreach my $initd_file (keys %initd_postinst) {
@@ -181,26 +200,26 @@ sub run {
                 $ok = 1 if $initd_path->link eq '/lib/init/upstart-job';
             }
             if (not $ok) {
-                tag 'init.d-script-not-included-in-package',
-                  "etc/init.d/$initd_file";
+                $self->tag('init.d-script-not-included-in-package',
+                    "etc/init.d/$initd_file");
             }
             next;
         }
 
         if (
             not $initd_path
-            or (    not $info->is_conffile($initd_path->name)
+            or (    not $processable->is_conffile($initd_path->name)
                 and not $initd_path->is_symlink)
-          ) {
-            tag 'init.d-script-not-marked-as-conffile',
-              "etc/init.d/$initd_file";
+        ) {
+            $self->tag('init.d-script-not-marked-as-conffile',
+                "etc/init.d/$initd_file");
         }
 
         # Check if file exists in package and check the script for
         # other issues if it was included in the package.
-        check_init($initd_path, $info);
+        $self->check_init($initd_path);
     }
-    check_defaults($info);
+    $self->check_defaults;
 
     return unless $initd_dir and $initd_dir->is_dir;
 
@@ -208,7 +227,11 @@ sub run {
     for my $script ($initd_dir->children) {
         my $tagname = 'script-in-etc-init.d-not-registered-via-update-rc.d';
         my $basename = $script->basename;
-        next if any {$basename eq $_} qw(README skeleton rc rcS);
+        next
+          if (
+            any {$basename eq $_}
+            qw(README skeleton rc rcS)
+          ) && !$script->is_dir;
 
         # In an upstart system, such as Ubuntu, init scripts are symlinks to
         # upstart-job which are not registered with update-rc.d.
@@ -222,8 +245,8 @@ sub run {
         # unregistered scripts so that we get more complete Lintian
         # coverage in the first pass.
         unless ($initd_postinst{$script->basename}) {
-            tag $tagname, $script;
-            check_init($script, $info);
+            $self->tag($tagname, $script);
+            $self->check_init($script);
         }
     }
 
@@ -231,7 +254,9 @@ sub run {
 }
 
 sub check_init {
-    my ($initd_path, $info) = @_;
+    my ($self, $initd_path) = @_;
+
+    my $processable = $self->processable;
 
     # In an upstart system, such as Ubuntu, init scripts are symlinks to
     # upstart-job.  It doesn't make sense to check the syntax of upstart-job,
@@ -245,24 +270,28 @@ sub check_init {
     my (%tag, %lsb);
     my $in_file_test = 0;
     my $needs_fs = 0;
-    my $fd = $initd_path->open;
+    open(my $fd, '<', $initd_path->unpacked_path);
     while (my $l = <$fd>) {
         if ($. == 1) {
-            if ($l =~ m,^\#!\s*(/usr/[^\s]+),) {
-                tag 'init.d-script-uses-usr-interpreter',$initd_path, $1;
-            } elsif ($l =~ m{^ [#]! \s* /lib/init/init-d-script}xsm) {
+            if ($l
+                =~ m{^ [#]! \s* (?:/usr/bin/env)? \s* /lib/init/init-d-script}xsm
+            ) {
                 for my $arg (qw(start stop restart force-reload status)) {
                     $tag{$arg} = 1;
                 }
+            } elsif ($l =~ m,^\#!\s*(/usr/[^\s]+),) {
+                $self->tag('init.d-script-uses-usr-interpreter',
+                    $initd_path, $1);
             }
         }
         if ($l =~ m/Please remove the "Author" lines|Example initscript/) {
-            tag 'init.d-script-contains-skeleton-template-content',
-              "${initd_path}:$.";
+            $self->tag('init.d-script-contains-skeleton-template-content',
+                "${initd_path}:$.");
         }
         if ($l =~ m/^\#\#\# BEGIN INIT INFO/) {
             if ($lsb{BEGIN}) {
-                tag 'init.d-script-has-duplicate-lsb-section',$initd_path;
+                $self->tag('init.d-script-has-duplicate-lsb-section',
+                    $initd_path);
                 next;
             }
             $lsb{BEGIN} = 1;
@@ -275,17 +304,19 @@ sub check_init {
                     $lsb{END} = 1;
                     last;
                 } elsif ($l !~ /^\#/) {
-                    tag 'init.d-script-has-unterminated-lsb-section',
-                      "${initd_path}:$.";
+                    $self->tag('init.d-script-has-unterminated-lsb-section',
+                        "${initd_path}:$.");
                     last;
                 } elsif ($l =~ /^\# ([a-zA-Z-]+):\s*(.*?)\s*$/) {
                     my $keyword = lc $1;
                     my $value = $2;
-                    tag 'init.d-script-has-duplicate-lsb-keyword',
-                      "${initd_path}:$. $keyword"
+                    $self->tag('init.d-script-has-duplicate-lsb-keyword',
+                        "${initd_path}:$. $keyword")
                       if (defined $lsb{$keyword});
-                    tag 'init.d-script-has-unknown-lsb-keyword',
-                      "${initd_path}:$. $keyword"
+                    $self->tag(
+                        'init.d-script-has-unknown-lsb-keyword',
+                        "${initd_path}:$. $keyword"
+                      )
                       unless (defined($lsb_keywords{$keyword})
                         || $keyword =~ /^x-/);
                     $lsb{$keyword} = defined($value) ? $value : '';
@@ -295,7 +326,8 @@ sub check_init {
                     $value =~ s/^\#\s*//;
                     $lsb{description} .= ' ' . $value;
                 } else {
-                    tag 'init.d-script-has-bad-lsb-line',"${initd_path}:$.";
+                    $self->tag('init.d-script-has-bad-lsb-line',
+                        "${initd_path}:$.");
                 }
             }
         }
@@ -307,7 +339,8 @@ sub check_init {
         $in_file_test = 0 if ($l =~ m/\bfi\b/);
         if (!$in_file_test && $l =~ m,^\s*\.\s+["'"]?(/etc/default/[\$\w/-]+),)
         {
-            tag 'init.d-script-sourcing-without-test',"${initd_path}:$. $1";
+            $self->tag('init.d-script-sourcing-without-test',
+                "${initd_path}:$. $1");
         }
 
         if ($l =~ m{\. /lib/init/init-d-script}) {
@@ -322,33 +355,36 @@ sub check_init {
         # text and the arguments to echo, etc.
         $needs_fs = 1 if ($l =~ m,^[^\#]*/var/,);
 
-        while ($l =~ s/^[^\#]*?(start|stop|restart|force-reload|status)//o) {
+        while ($l =~ s/^[^\#]*?(start|stop|restart|force-reload|status)//) {
             $tag{$1} = 1;
         }
 
-        if (   $l =~ m{^\s*\.\s+/lib/lsb/init-functions}
-            && !$info->relation('strong')->implies('lsb-base')
-            && none { $_->basename =~ m/\.service$/ } $info->sorted_index) {
-            tag 'init.d-script-needs-depends-on-lsb-base',
-              $initd_path, "(line $.)";
+        if (
+               $l =~ m{^\s*\.\s+/lib/lsb/init-functions}
+            && !$processable->relation('strong')->implies('lsb-base')
+            && none { $_->basename =~ m/\.service$/ && !$_->is_dir }
+            $processable->installed->sorted_list
+        ) {
+            $self->tag('init.d-script-needs-depends-on-lsb-base',
+                $initd_path, "(line $.)");
         }
     }
     close($fd);
 
     # Make sure all of the required keywords are present.
     if (not $lsb{BEGIN}) {
-        tag 'init.d-script-missing-lsb-section', $initd_path;
+        $self->tag('init.d-script-missing-lsb-section', $initd_path);
     } else {
         for my $keyword (keys %lsb_keywords) {
             if ($lsb_keywords{$keyword} && !defined $lsb{$keyword}) {
                 if ($keyword eq 'short-description') {
-                    tag 'init.d-script-missing-lsb-short-description',
-                      $initd_path;
+                    $self->tag('init.d-script-missing-lsb-short-description',
+                        $initd_path);
                 } elsif ($keyword eq 'description') {
                     next;
                 } else {
-                    tag 'init.d-script-missing-lsb-keyword',
-                      $initd_path, $keyword;
+                    $self->tag('init.d-script-missing-lsb-keyword',
+                        $initd_path, $keyword);
                 }
             }
         }
@@ -361,12 +397,12 @@ sub check_init {
             if ($runlevel =~ /^[sS0-6]$/) {
                 $start{lc $runlevel} = 1;
                 if ($runlevel eq '0' or $runlevel eq '6') {
-                    tag 'init.d-script-starts-in-stop-runlevel',
-                      $initd_path, $runlevel;
+                    $self->tag('init.d-script-starts-in-stop-runlevel',
+                        $initd_path, $runlevel);
                 }
             } else {
-                tag 'init.d-script-has-bad-start-runlevel',
-                  $initd_path, $runlevel;
+                $self->tag('init.d-script-has-bad-start-runlevel',
+                    $initd_path, $runlevel);
             }
         }
 
@@ -375,7 +411,7 @@ sub check_init {
         my $start = join(' ', sort grep {$_ =~ /^[2-5]$/} keys %start);
         if (length($start) > 0 and $start ne '2 3 4 5') {
             my @missing = grep { !defined $start{$_} } qw(2 3 4 5);
-            tag 'init.d-script-missing-start', $initd_path,@missing;
+            $self->tag('init.d-script-missing-start', $initd_path,@missing);
         }
     }
     if (defined $lsb{'default-stop'}) {
@@ -384,15 +420,16 @@ sub check_init {
             if ($runlevel =~ /^[sS0-6]$/) {
                 $stop{$runlevel} = 1 unless $runlevel =~ /[sS2-5]/;
                 if ($start{$runlevel}) {
-                    tag 'init.d-script-has-conflicting-start-stop',
-                      $initd_path, $runlevel;
+                    $self->tag('init.d-script-has-conflicting-start-stop',
+                        $initd_path, $runlevel);
                 }
                 if ($runlevel =~ /[sS]/) {
-                    tag 'init-d-script-stops-in-s-runlevel',$initd_path;
+                    $self->tag('init-d-script-stops-in-s-runlevel',
+                        $initd_path);
                 }
             } else {
-                tag 'init.d-script-has-bad-stop-runlevel',
-                  $initd_path, $runlevel;
+                $self->tag('init.d-script-has-bad-stop-runlevel',
+                    $initd_path, $runlevel);
             }
         }
 
@@ -403,7 +440,10 @@ sub check_init {
             my $base = $initd_path->basename;
             if (none { $base eq $_ } qw(killprocs sendsigs halt reboot)) {
                 my @missing = grep { !defined $stop{$_} } qw(0 1 6);
-                tag 'init.d-script-possible-missing-stop',$initd_path,@missing;
+                my $start = join(' ', sort keys %start);
+                $self->tag('init.d-script-possible-missing-stop',
+                    $initd_path,@missing)
+                  unless $start eq 's';
             }
         }
     }
@@ -411,14 +451,14 @@ sub check_init {
         my $provides_self;
         for my $facility (split(/\s+/, $lsb{'provides'})) {
             if ($facility =~ /^\$/) {
-                tag 'init.d-script-provides-virtual-facility',
-                  $initd_path, $facility;
+                $self->tag('init.d-script-provides-virtual-facility',
+                    $initd_path, $facility);
             }
             if ($initd_path->basename =~/^\Q$facility\E(?:.sh)?$/) {
                 $provides_self = 1;
             }
         }
-        tag 'init.d-script-does-not-provide-itself', $initd_path
+        $self->tag('init.d-script-does-not-provide-itself', $initd_path)
           unless $provides_self;
     }
 
@@ -430,8 +470,8 @@ sub check_init {
         my @required = split(' ', $lsb{'required-start'} || '');
         if ($needs_fs) {
             if (none { /^\$(?:local_fs|remote_fs|all)\z/ } @required) {
-                tag 'init.d-script-missing-dependency-on-local_fs',
-                  "${initd_path}: required-start";
+                $self->tag('init.d-script-missing-dependency-on-local_fs',
+                    "${initd_path}: required-start");
             }
         }
     }
@@ -441,9 +481,9 @@ sub check_init {
             if (
                 none { /^(?:\$(?:local|remote)_fs|\$all|umountn?fs)\z/ }
                 @required
-              ) {
-                tag 'init.d-script-missing-dependency-on-local_fs',
-                  "${initd_path}: required-stop";
+            ) {
+                $self->tag('init.d-script-missing-dependency-on-local_fs',
+                    "${initd_path}: required-stop");
             }
         }
     }
@@ -454,17 +494,17 @@ sub check_init {
         next unless defined $lsb{$keyword};
         for my $dependency (split(/\s+/, $lsb{$keyword})) {
             if (defined $implied_dependencies{$dependency}) {
-                tag 'init.d-script-should-depend-on-virtual-facility',
-                  $initd_path,
-                  "$dependency -> $implied_dependencies{$dependency}";
+                $self->tag('non-virtual-facility-in-initd-script',
+                    $initd_path,
+                    "$dependency -> $implied_dependencies{$dependency}");
             } elsif ($keyword =~ m/^required-/ && $dependency =~ m/^\$/) {
-                tag 'init.d-script-depends-on-unknown-virtual-facility',
-                  $initd_path, $dependency
+                $self->tag('init.d-script-depends-on-unknown-virtual-facility',
+                    $initd_path, $dependency)
                   unless ($VIRTUAL_FACILITIES->known($dependency));
             }
             if ($dependency =~ m/^\$all$/) {
-                tag 'init.d-script-depends-on-all-virtual-facility',
-                  $initd_path, $keyword;
+                $self->tag('init.d-script-depends-on-all-virtual-facility',
+                    $initd_path, $keyword);
             }
         }
     }
@@ -472,35 +512,70 @@ sub check_init {
     # all tags included in file?
     for my $option (qw(start stop restart force-reload)) {
         $tag{$option}
-          or tag 'init.d-script-does-not-implement-required-option',
-          $initd_path, $option;
+          or $self->tag('init.d-script-does-not-implement-required-option',
+            $initd_path, $option);
     }
 
-    for my $option (qw(status)) {
-        $tag{$option}
-          or tag 'init.d-script-does-not-implement-optional-option',
-          $initd_path, $option;
-    }
+    $self->tag('init.d-script-does-not-implement-status-option', $initd_path)
+      unless $tag{'status'};
 
     return;
 }
 
 sub check_defaults {
-    my ($info) = @_;
-    my $dir = $info->index_resolved_path('etc/default/');
+    my ($self) = @_;
+
+    my $processable = $self->processable;
+
+    my $dir = $processable->installed->resolve_path('etc/default/');
     return unless $dir and $dir->is_dir;
     for my $path ($dir->children) {
         return if not $path->is_open_ok;
-        my $fd = $path->open;
+        open(my $fd, '<', $path->unpacked_path);
         while (<$fd>) {
-            tag 'init.d-script-should-always-start-service', $path, "(line $.)"
-              if m/^\s*#*\s*(?:ENABLED|DISABLED|[A-Z]*RUN)=/;
+            $self->tag('init.d-script-should-always-start-service',
+                $path, "(line $.)")
+              if m/$ALWAYS_START/;
         }
         close($fd);
     }
     return;
 }
 
+sub visit_installed_files {
+    my ($self, $file) = @_;
+
+    # check for missing init.d script when alternative init system is present
+
+    if (   $file =~ m,etc/sv/([^/]+)/run$,
+        or $file =~ m,lib/systemd/system/([^/@]+)\.service,) {
+
+        my $service = $1;
+
+        $self->tag('package-supports-alternative-init-but-no-init.d-script',
+            $file)
+          unless $self->processable->installed->resolve_path(
+            "etc/init.d/${service}")
+          or $self->processable->installed->resolve_path(
+            "lib/systemd/system/${service}.path")
+          or $self->processable->installed->resolve_path(
+            "lib/systemd/system/${service}.timer");
+    }
+
+    if ($file =~ m,etc/sv/([^/]+)/$,) {
+        my $service = $1;
+        my $file
+          = $self->processable->installed->resolve_path(
+            "etc/sv/${service}/run");
+        $self->tag(
+            'directory-in-etc-sv-directory-without-executable-run-script',
+            $file)
+          if not $file or not $file->is_executable;
+    }
+
+    return;
+}
+
 1;
 
 # Local Variables:
